@@ -14,7 +14,7 @@ import secrets
 import struct
 import time
 from random import randint
-from typing import Any, Callable, Coroutine, Mapping, Optional
+from typing import Any, Callable, Coroutine, Optional
 
 import aiohttp
 from Crypto.Cipher import AES
@@ -37,13 +37,22 @@ from .containers import (
     UserData,
     WashTowelMode,
 )
-from .exceptions import RoborockException, RoborockTimeout, VacuumError
+from .exceptions import (
+    RoborockAccountDoesNotExist,
+    RoborockException,
+    RoborockInvalidCode,
+    RoborockInvalidEmail,
+    RoborockTimeout,
+    RoborockUrlException,
+    VacuumError,
+)
 from .roborock_future import RoborockFuture
 from .roborock_message import RoborockMessage
 from .typing import DeviceProp, DockSummary, RoborockCommand
 from .util import unpack_list
 
 _LOGGER = logging.getLogger(__name__)
+KEEPALIVE = 60
 QUEUE_TIMEOUT = 4
 SPECIAL_COMMANDS = [
     RoborockCommand.GET_MAP_V1,
@@ -76,21 +85,30 @@ class PreparedRequest:
 
 
 class RoborockClient:
-    def __init__(self, endpoint: str, devices_info: Mapping[str, RoborockDeviceInfo]) -> None:
-        self.devices_info = devices_info
+    def __init__(self, endpoint: str, device_info: RoborockDeviceInfo) -> None:
+        self.device_info = device_info
         self._endpoint = endpoint
         self._nonce = secrets.token_bytes(16)
         self._waiting_queue: dict[int, RoborockFuture] = {}
-        self._status_listeners: list[Callable[[int, str], None]] = []
+        self._last_device_msg_in = self.time_func()
+        self._last_disconnection = self.time_func()
+        self.keep_alive = KEEPALIVE
 
-    def add_status_listener(self, callback: Callable[[int, str], None]):
-        self._status_listeners.append(callback)
+    @property
+    def time_func(self) -> Callable[[], float]:
+        try:
+            # Use monotonic clock if available
+            time_func = time.monotonic
+        except AttributeError:
+            time_func = time.time
+        return time_func
 
     async def async_disconnect(self) -> Any:
         raise NotImplementedError
 
     def on_message(self, messages: list[RoborockMessage]) -> None:
         try:
+            self._last_device_msg_in = self.time_func()
             for data in messages:
                 protocol = data.protocol
                 if protocol == 102 or protocol == 4:
@@ -134,6 +152,19 @@ class RoborockClient:
         except Exception as ex:
             _LOGGER.exception(ex)
 
+    def on_disconnect(self, exc: Optional[Exception]) -> None:
+        self._last_disconnection = self.time_func()
+        _LOGGER.warning("Roborock client disconnected")
+        if exc is not None:
+            _LOGGER.warning(exc)
+
+    def should_keepalive(self) -> bool:
+        now = self.time_func()
+        # noinspection PyUnresolvedReferences
+        if now - self._last_disconnection > self.keep_alive**2 and now - self._last_device_msg_in > self.keep_alive:
+            return False
+        return True
+
     async def _async_response(self, request_id: int, protocol_id: int = 0) -> tuple[Any, VacuumError | None]:
         try:
             queue = RoborockFuture(protocol_id)
@@ -169,27 +200,27 @@ class RoborockClient:
         )
         return request_id, timestamp, payload
 
-    async def send_command(self, device_id: str, method: RoborockCommand, params: Optional[list] = None):
+    async def send_command(self, method: RoborockCommand, params: Optional[list] = None):
         raise NotImplementedError
 
-    async def get_status(self, device_id: str) -> Status | None:
-        status = await self.send_command(device_id, RoborockCommand.GET_STATUS)
+    async def get_status(self) -> Status | None:
+        status = await self.send_command(RoborockCommand.GET_STATUS)
         if isinstance(status, dict):
             return Status.from_dict(status)
         return None
 
-    async def get_dnd_timer(self, device_id: str) -> DNDTimer | None:
+    async def get_dnd_timer(self) -> DNDTimer | None:
         try:
-            dnd_timer = await self.send_command(device_id, RoborockCommand.GET_DND_TIMER)
+            dnd_timer = await self.send_command(RoborockCommand.GET_DND_TIMER)
             if isinstance(dnd_timer, dict):
                 return DNDTimer.from_dict(dnd_timer)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_clean_summary(self, device_id: str) -> CleanSummary | None:
+    async def get_clean_summary(self) -> CleanSummary | None:
         try:
-            clean_summary = await self.send_command(device_id, RoborockCommand.GET_CLEAN_SUMMARY)
+            clean_summary = await self.send_command(RoborockCommand.GET_CLEAN_SUMMARY)
             if isinstance(clean_summary, dict):
                 return CleanSummary.from_dict(clean_summary)
             elif isinstance(clean_summary, list):
@@ -201,55 +232,54 @@ class RoborockClient:
             _LOGGER.error(e)
         return None
 
-    async def get_clean_record(self, device_id: str, record_id: int) -> CleanRecord | None:
+    async def get_clean_record(self, record_id: int) -> CleanRecord | None:
         try:
-            clean_record = await self.send_command(device_id, RoborockCommand.GET_CLEAN_RECORD, [record_id])
+            clean_record = await self.send_command(RoborockCommand.GET_CLEAN_RECORD, [record_id])
             if isinstance(clean_record, dict):
                 return CleanRecord.from_dict(clean_record)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_consumable(self, device_id: str) -> Consumable | None:
+    async def get_consumable(self) -> Consumable | None:
         try:
-            consumable = await self.send_command(device_id, RoborockCommand.GET_CONSUMABLE)
+            consumable = await self.send_command(RoborockCommand.GET_CONSUMABLE)
             if isinstance(consumable, dict):
                 return Consumable.from_dict(consumable)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_wash_towel_mode(self, device_id: str) -> WashTowelMode | None:
+    async def get_wash_towel_mode(self) -> WashTowelMode | None:
         try:
-            washing_mode = await self.send_command(device_id, RoborockCommand.GET_WASH_TOWEL_MODE)
+            washing_mode = await self.send_command(RoborockCommand.GET_WASH_TOWEL_MODE)
             if isinstance(washing_mode, dict):
                 return WashTowelMode.from_dict(washing_mode)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_dust_collection_mode(self, device_id: str) -> DustCollectionMode | None:
+    async def get_dust_collection_mode(self) -> DustCollectionMode | None:
         try:
-            dust_collection = await self.send_command(device_id, RoborockCommand.GET_DUST_COLLECTION_MODE)
+            dust_collection = await self.send_command(RoborockCommand.GET_DUST_COLLECTION_MODE)
             if isinstance(dust_collection, dict):
                 return DustCollectionMode.from_dict(dust_collection)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_smart_wash_params(self, device_id: str) -> SmartWashParams | None:
+    async def get_smart_wash_params(self) -> SmartWashParams | None:
         try:
-            mop_wash_mode = await self.send_command(device_id, RoborockCommand.GET_SMART_WASH_PARAMS)
+            mop_wash_mode = await self.send_command(RoborockCommand.GET_SMART_WASH_PARAMS)
             if isinstance(mop_wash_mode, dict):
                 return SmartWashParams.from_dict(mop_wash_mode)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_dock_summary(self, device_id: str, dock_type: RoborockEnum) -> DockSummary | None:
+    async def get_dock_summary(self, dock_type: RoborockEnum) -> DockSummary | None:
         """Gets the status summary from the dock with the methods available for a given dock.
 
-        :param device_id: Device id
         :param dock_type: RoborockDockTypeCode"""
         try:
             commands: list[
@@ -258,11 +288,11 @@ class RoborockClient:
                     Any,
                     DustCollectionMode | WashTowelMode | SmartWashParams | None,
                 ]
-            ] = [self.get_dust_collection_mode(device_id)]
+            ] = [self.get_dust_collection_mode()]
             if dock_type == RoborockDockTypeCode["3"]:
                 commands += [
-                    self.get_wash_towel_mode(device_id),
-                    self.get_smart_wash_params(device_id),
+                    self.get_wash_towel_mode(),
+                    self.get_smart_wash_params(),
                 ]
             [dust_collection_mode, wash_towel_mode, smart_wash_params] = unpack_list(
                 list(await asyncio.gather(*commands)), 3
@@ -273,21 +303,21 @@ class RoborockClient:
             _LOGGER.error(e)
         return None
 
-    async def get_prop(self, device_id: str) -> DeviceProp | None:
+    async def get_prop(self) -> DeviceProp | None:
         [status, dnd_timer, clean_summary, consumable] = await asyncio.gather(
             *[
-                self.get_status(device_id),
-                self.get_dnd_timer(device_id),
-                self.get_clean_summary(device_id),
-                self.get_consumable(device_id),
+                self.get_status(),
+                self.get_dnd_timer(),
+                self.get_clean_summary(),
+                self.get_consumable(),
             ]
         )
         last_clean_record = None
         if clean_summary and clean_summary.records and len(clean_summary.records) > 0:
-            last_clean_record = await self.get_clean_record(device_id, clean_summary.records[0])
+            last_clean_record = await self.get_clean_record(clean_summary.records[0])
         dock_summary = None
         if status and status.dock_type is not None and status.dock_type != RoborockDockTypeCode["0"]:
-            dock_summary = await self.get_dock_summary(device_id, status.dock_type)
+            dock_summary = await self.get_dock_summary(status.dock_type)
         if any([status, dnd_timer, clean_summary, consumable]):
             return DeviceProp(
                 status,
@@ -299,27 +329,27 @@ class RoborockClient:
             )
         return None
 
-    async def get_multi_maps_list(self, device_id) -> MultiMapsList | None:
+    async def get_multi_maps_list(self) -> MultiMapsList | None:
         try:
-            multi_maps_list = await self.send_command(device_id, RoborockCommand.GET_MULTI_MAPS_LIST)
+            multi_maps_list = await self.send_command(RoborockCommand.GET_MULTI_MAPS_LIST)
             if isinstance(multi_maps_list, dict):
                 return MultiMapsList.from_dict(multi_maps_list)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_networking(self, device_id) -> NetworkInfo | None:
+    async def get_networking(self) -> NetworkInfo | None:
         try:
-            networking_info = await self.send_command(device_id, RoborockCommand.GET_NETWORK_INFO)
+            networking_info = await self.send_command(RoborockCommand.GET_NETWORK_INFO)
             if isinstance(networking_info, dict):
                 return NetworkInfo.from_dict(networking_info)
         except RoborockTimeout as e:
             _LOGGER.error(e)
         return None
 
-    async def get_room_mapping(self, device_id: str) -> list[RoomMapping]:
+    async def get_room_mapping(self) -> list[RoomMapping]:
         """Gets the mapping from segment id -> iot id. Only works on local api."""
-        mapping = await self.send_command(device_id, RoborockCommand.GET_ROOM_MAPPING)
+        mapping = await self.send_command(RoborockCommand.GET_ROOM_MAPPING)
         if isinstance(mapping, list):
             return [
                 RoomMapping(segment_id=segment_id, iot_id=iot_id)  # type: ignore
@@ -345,12 +375,15 @@ class RoborockApiClient:
                 params={"email": self._username, "needtwostepauth": "false"},
             )
             if response is None:
-                raise RoborockException("get url by email returned None")
-            if response.get("code") != 200:
-                raise RoborockException(response.get("error"))
+                raise RoborockUrlException("get url by email returned None")
+            response_code = response.get("code")
+            if response_code != 200:
+                if response_code == 2003:
+                    raise RoborockInvalidEmail("Your email was incorrectly formatted.")
+                raise RoborockUrlException(response.get("error"))
             response_data = response.get("data")
             if response_data is None:
-                raise RoborockException("response does not have 'data'")
+                raise RoborockUrlException("response does not have 'data'")
             self.base_url = response_data.get("url")
         return self.base_url
 
@@ -375,8 +408,12 @@ class RoborockApiClient:
         )
         if code_response is None:
             raise RoborockException("Failed to get a response from send email code")
-        if code_response.get("code") != 200:
-            raise RoborockException(code_response.get("msg"))
+        response_code = code_response.get("code")
+        if response_code != 200:
+            if response_code == 2008:
+                raise RoborockAccountDoesNotExist("Account does not exist - check your login and try again.")
+            else:
+                raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")
 
     async def pass_login(self, password: str) -> UserData:
         base_url = await self._get_base_url()
@@ -395,7 +432,7 @@ class RoborockApiClient:
         if login_response is None:
             raise RoborockException("Login response is none")
         if login_response.get("code") != 200:
-            raise RoborockException(login_response.get("msg"))
+            raise RoborockException(f"{login_response.get('msg')} - response code: {login_response.get('code')}")
         user_data = login_response.get("data")
         if not isinstance(user_data, dict):
             raise RoborockException("Got unexpected data type for user_data")
@@ -417,8 +454,11 @@ class RoborockApiClient:
         )
         if login_response is None:
             raise RoborockException("Login request response is None")
-        if login_response.get("code") != 200:
-            raise RoborockException(login_response.get("msg"))
+        response_code = login_response.get("code")
+        if response_code != 200:
+            if response_code == 2018:
+                raise RoborockInvalidCode("Invalid code - check your code and try again.")
+            raise RoborockException(f"{login_response.get('msg')} - response code: {response_code}")
         user_data = login_response.get("data")
         if not isinstance(user_data, dict):
             raise RoborockException("Got unexpected data type for user_data")
@@ -439,7 +479,7 @@ class RoborockApiClient:
         if home_id_response is None:
             raise RoborockException("home_id_response is None")
         if home_id_response.get("code") != 200:
-            raise RoborockException(home_id_response.get("msg"))
+            raise RoborockException(f"{home_id_response.get('msg')} - response code: {home_id_response.get('code')}")
 
         home_id = home_id_response["data"].get("rrHomeId")
         timestamp = math.floor(time.time())
