@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import json
 import logging
 from pathlib import Path
@@ -16,21 +18,21 @@ _LOGGER = logging.getLogger(__name__)
 
 class RoborockContext:
     roborock_file = Path("~/.roborock").expanduser()
-    _login_data: LoginData = None
+    _login_data: LoginData | None = None
 
     def __init__(self):
         self.reload()
 
     def reload(self):
         if self.roborock_file.is_file():
-            with open(self.roborock_file, 'r') as f:
+            with open(self.roborock_file, "r") as f:
                 data = json.load(f)
                 if data:
                     self._login_data = LoginData.from_dict(data)
 
     def update(self, login_data: LoginData):
         data = json.dumps(login_data.as_dict(), default=vars)
-        with open(self.roborock_file, 'w') as f:
+        with open(self.roborock_file, "w") as f:
             f.write(data)
         self.reload()
 
@@ -48,9 +50,7 @@ class RoborockContext:
 @click.group()
 @click.pass_context
 def cli(ctx, debug: int):
-    logging_config: Dict[str, Any] = {
-        "level": logging.DEBUG if debug > 0 else logging.INFO
-    }
+    logging_config: Dict[str, Any] = {"level": logging.DEBUG if debug > 0 else logging.INFO}
     logging.basicConfig(**logging_config)  # type: ignore
     ctx.obj = RoborockContext()
 
@@ -81,7 +81,8 @@ async def _discover(ctx):
     home_data = await client.get_home_data(login_data.user_data)
     context.update(LoginData({**login_data, "home_data": home_data}))
     click.echo(
-        f"Discovered devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}")
+        f"Discovered devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}"
+    )
 
 
 @click.command()
@@ -101,26 +102,30 @@ async def list_devices(ctx):
         await _discover(ctx)
         login_data = context.login_data()
     home_data = login_data.home_data
-    click.echo(f"Known devices {', '.join([device.name for device in home_data.devices + home_data.received_devices])}")
+    device_name_id = ", ".join(
+        [f"{device.name}: {device.duid}" for device in home_data.devices + home_data.received_devices]
+    )
+    click.echo(f"Known devices {device_name_id}")
 
 
 @click.command()
+@click.option("--device_id", required=True)
 @click.option("--cmd", required=True)
 @click.option("--params", required=False)
 @click.pass_context
 @run_sync()
-async def command(ctx, cmd, params):
+async def command(ctx, cmd, device_id, params):
     context: RoborockContext = ctx.obj
     login_data = context.login_data()
     if not login_data.home_data:
         await _discover(ctx)
         login_data = context.login_data()
     home_data = login_data.home_data
-    device_map: dict[str, RoborockDeviceInfo] = {}
-    for device in home_data.devices + home_data.received_devices:
-        device_map[device.duid] = RoborockDeviceInfo(device=device)
-    mqtt_client = RoborockMqttClient(login_data.user_data, device_map)
-    await mqtt_client.send_command(home_data.devices[0].duid, cmd, params)
+    devices = home_data.devices + home_data.received_devices
+    device = next((device for device in devices if device.duid == device_id), None)
+    device_info = RoborockDeviceInfo(device=device)
+    mqtt_client = RoborockMqttClient(login_data.user_data, device_info)
+    await mqtt_client.send_command(cmd, params)
     mqtt_client.__del__()
 
 
